use picodata_sdk::*;
use tarantool::fiber;
use tarantool::fiber::r#async::watch;

use crate::instance::GradeVariant;
use crate::loop_start;
use crate::r#loop::FlowControl;
use crate::tlog;

static mut PLUGIN_MANAGER: Option<PluginManager> = None;

pub struct PluginManager {
    _loop: Option<fiber::JoinHandle<'static, ()>>,
    loop_waker: watch::Sender<GradeVariant>,

    plugin_list: Vec<Plugin>,
}

struct PluginManagerState {
    loop_waker: watch::Receiver<GradeVariant>,
}

impl PluginManager {
    pub fn init(plugins: &[String]) {
        assert!(unsafe { &PLUGIN_MANAGER }.is_none());

        let (loop_waker_tx, loop_waker_rx) = watch::channel(GradeVariant::default());
        let state = PluginManagerState {
            loop_waker: loop_waker_rx,
        };
        let plugin_manager = PluginManager {
            _loop: loop_start!("plugin_manager_loop", Self::iter_fn, state),
            loop_waker: loop_waker_tx,
            plugin_list: Self::plug_list_init(plugins),
        };

        unsafe { PLUGIN_MANAGER = Some(plugin_manager) };
    }

    async fn iter_fn(state: &mut PluginManagerState) -> FlowControl {
        if state.loop_waker.changed().await.is_ok() {
            println!("Invoke the plugin callback for GradeChanged event here");

            let variant = state.loop_waker.borrow();
            match *variant {
                GradeVariant::Offline => {
                    println!("Invoke the plugin callback for {} here", *variant)
                }
                GradeVariant::Replicated => {
                    println!("Invoke the plugin callback for {} here", *variant)
                }
                GradeVariant::ShardingInitialized => {
                    println!("Invoke the plugin callback for {} here", *variant)
                }
                GradeVariant::Online => {
                    println!("Invoke the plugin callback for {} here", *variant);
                    Self::online();
                }
                GradeVariant::Expelled => {
                    println!("Invoke the plugin callback for {} here", *variant)
                }
            }
        }

        FlowControl::Continue
    }

    fn plug_list_init(plugins: &[String]) -> Vec<Plugin> {
        let mut plugin_list: Vec<Plugin> = Vec::new();
        plugins.iter().for_each(|path| {
            if let Some(plugin) = Plugin::new().load(path) {
                plugin_list.push(plugin);
            }
        });
        plugin_list.iter().for_each(|plugin| {
            tlog!(Info, "Plugin: '{}' loaded", plugin.name());
        });

        plugin_list
    }

    pub fn notify_grade_changed(variant: GradeVariant) {
        if let Some(plugin_manager_loop) = unsafe { &PLUGIN_MANAGER } {
            let _ = plugin_manager_loop.loop_waker.send(variant);
        }
    }

    pub fn online() {
        if let Some(plugin_manager_loop) = unsafe { &PLUGIN_MANAGER } {
            plugin_manager_loop.plugin_list.iter().for_each(|plugin| {
                plugin.online();
            });
        }
    }

    pub fn deinit() {
        if let Some(plugin_manager_loop) = unsafe { &PLUGIN_MANAGER } {
            plugin_manager_loop.plugin_list.iter().for_each(|plugin| {
                plugin.shutdown();
            });
        }
    }
}
